#!/usr/bin/perl -w
#
# quizz CGI script
#
# Thomas Linden <tom@co.daemon.de>
#
# Copyright (c) 2003 Thomas Linden.
# ALL RIGHTS RESERVED.
#
# Licensed under the terms of the artistic license.
#
# The module "Template" is Copyright 1999-2000 Jens Heunemann.
# Used with permission.
#


use CGI qw/:standard/;
use Data::Dumper;
use CGI::Carp qw(fatalsToBrowser);
use CGI::Cookie;
use Template;
use Config::General;
use GDBM_File;
use Digest::MD5 qw(md5_base64);
use Fcntl qw(:flock);

# initialize vars, change as appropriate
my $cgi            = "";
my $path_conf      = "config";
my $path_data      = "data";
my $path_templates = "templates";


# read/parse config file using Config::General
my %config = ParseConfig(
			 -ConfigFile       => "$path_conf/quizz.cfg",
			 -UseApacheInclude => 1,
			 -IncludeRelative  => 1
			   );

my %choice = (
	      M => "multiple",
	      O => "multiple",
	      S => "single",
	      R => "regex"
	      );

my $VERSION  = "1.2.0";

my @loop = @{$config{loop}};

# cgi object
my $query = new CGI;
my $now   = time;

# result+stats databases (gdbm)
my (%db, %stats) ;

# initialize if not already done so far
&prepare;

# initialize shared templates
my $head     = new Template("$path_templates/header.tpl");
my $foot     = new Template("$path_templates/footer.tpl");

# initialize cgi parameters
my %input = map { $_ => $query->param($_) } $query->param;

# check it up where to go
my $sub2go   = $input{state} || "default";




eval {
  # now do the real work
  &$sub2go;
};
if ($@) {
  print "Content-Type: text/html\n\n\n";
  print $@;
}
&dump(\%input);

# and the foot including the last changed date
my @stat = lstat($0);
$foot->code(
	    date => scalar localtime( $stat[9] ),
	    version => $VERSION
	   );
$foot->print;

exit;





###### all subs from now on, called by ::main via closure call (&$sub2go) #####
sub stats {
  print "Content-Type: text/html\n\n\n";
  $head->print;
  &stats_out;
}

sub default {
  print "Content-Type: text/html\n\n\n";
  $head->print;
  my $T = new Template(-file => "$path_templates/default.tpl");

  # compute best-of list
  my @_bestof;
  foreach my $user (keys %db) {
    my %entry = &get_entry($user);
    push @_bestof, \%entry;
  }
  my $num = 1;
  foreach my $entry (sort { $b->{points} <=>  $a->{points} } @_bestof) {
    if (($num <= $config{config}->{max_best}) || $input{showall}) {
      $entry->{pos}   = $num;
      $entry->{color} = $loop[0];
      $entry->{cgi}   = $cgi;
      @loop = reverse @loop;
      push @bestof, $entry;
    }
    else {
      last;
    }
    $num++;
  }

  $T->code(
	   data => {
		    quiz_name        => $config{quiz_name},
		    quiz_description => $config{quiz_description},
		    cgi              => $cgi
		   },
	   bestof => \@bestof
	  );

  print $T->process;
}




sub check_cookie {
  if ($config{config}->{cookie} ne "none") {
    # check cookie
    my %cookie = fetch CGI::Cookie;
    my ($name, $md5) = &get_md5;
    if (exists $cookie{$name}) {
      my $session = $cookie{$name}->value();
      if ($md5 == $session) {
	&default; # not allowed!
	print "<I>no retry as user <b>$input{user}</b> possible.</I>\n";
	return;
      }
    }
    else {
      # create new cookie
      my $c = new CGI::Cookie(
			      -name    => $name,
			      -value   => $md5,
			      -expires => $config{config}->{cookie_expire} || '+1M', # expires after 1 week
			     );
      print header(-cookie=>[$c]);
      $head->print;
    }
  }
  else {
    # no cookies, but dont forget to send the header
    print "Content-Type: text/html\n\n\n";
    $head->print;
  }
  return 1;
}





sub compute_question {
  if (! exists $input{question}) {
    # starting point
    $input{question} = 1;

    # remove eventually existing stats entry for this user
    delete $stats{$input{user}};

    # save personal data, if any
    my %userdat = map {$_ => $input{$_} } qw(real email uri);
    &add_entry(
	       time => 0,
	       points => 0,
	       user => $input{user},
	       date => scalar localtime(time),
	       %userdat
	      );

    # check cookies and continue
    return if(! &check_cookie );
  }
  else {
    # previous run was a question
    print "Content-Type: text/html\n\n\n";
    $head->print;

    my %q = %{ $config{question}->{ $input{question} } }; # current question
    my $factor = $q{points} / $q{time};                   # how to calculate time from points
    my %entry = &get_entry($input{user});                 # results entry for current user, if any
    my ($points, $time, $result, %stats);

    if ($input{question} == 1) {
      # reset
      $entry{points} = 0;
      $entry{time}   = 0;
    }

    # calculate time and points based on time.
    # points will be zeroed if answer is wrong (later)!
    #
    # $points and $time               - summaries
    # $input{points} and $input{time} - current results
    $input{time}   = $now - $input{starttime};
    $input{points} = $q{points} - int($input{time} * $factor);
    $time          = $input{time} + $entry{time};
    $points        = $input{points} + $entry{points};

    print "time needed: $input{time}<br>time required: $q{time}<br>" if($config{debug});

    if ($input{time} >= $q{time}) {
      # check for timeout
      $points = $entry{points};
      $input{points} = 0;
      $input{time}   = $now - $input{starttime};
      $result = "timeout";
    }
    else {
      # compute the answer
      $result = "correct"; # default
      if ($q{type} eq "M" || $q{type} eq "O") {
	# multiple/optional-multiple choice
	my %choices = map { $_ => 1 } $query->param("choice");
	my %correct = map { $_ => 1 } split /,/, $q{result};
	$input{given} = join ",", keys %choices;

        my $ncorrect = scalar keys %correct; # all possible correct answers

	foreach my $choice (keys %choices) {
	  if (exists $correct{$choice}) {
	    delete $correct{$choice};
	    delete $choices{$choice};
	  }
	}
	my $remaining = scalar keys %correct;      # not all possible were answered
	my $missing   = scalar keys %choices;      # wrong answers were choosen
        my $possible  = scalar keys %{$q{choice}};  # all possible choices
	if ($q{type} eq "M") {
	  if ($remaining || $missing) {
	    # shit, wrong answer!
	    $points = $entry{points};
	    $result = "wrong";
	    $input{points} = 0;
	  }
	}
	else {
	  # $q{type} eq "O"
	  # wrong if either a wrong answer were choosen
	  # or if not enough answers were given
	  print "possible: $possible, ncorrect: $ncorrect<br>\n" if($config{debug});
	  &dump($q{choice});
	  print "result: $q{result}, missing: $missing, remaining: $remaining, required: $q{required}\n" if($config{debug});
	  if ($missing || $remaining > $q{required}) {
	    # shit, wrong answer!
	    $points = $entry{points};
	    $result = "wrong";
	    $input{points} = 0;
	  }
	}
      }
      elsif ($q{type} eq "S") {
	# single choice
	$input{given} = $input{choice};
	if ($input{choice} ne $q{result}) {
	  # shit, wrong answer!
	  $points = $entry{points};
	  $result = "wrong";
	  $input{points} = 0;
	}
      }
      elsif ($q{type} eq "R") {
	# regex choice
	$input{given} = $input{choice};
	$input{given} =~ s/:/%%/g; # mask delimiter
	if ($input{choice} !~ /$q{result}/) {
	  # shit, wrong answer!
	  $points = $entry{points};
	  $result = "wrong";
	  $input{points} = 0;
	}
      }

      if ($result eq "correct") {
	if (exists $q{latency_limit}) {
	  if ($input{time} <= $q{latency_limit}) {
	    # within question latency limit
	    $input{points} = $q{points};
	    $points = $input{points} + $entry{points};
	  }
	}
	elsif (exists $config{config}->{latency_limit}) {
	  if ($input{time} <= $config{config}->{latency_limit}) {
	    # within global latency limit
	    $input{points} = $q{points};
	    $points = $input{points} + $entry{points};
	  }
	}
      }
    }

    print "result: $result<br>points: $input{points}" if($config{debug});

    my %userdat = map {$_ => $entry{$_} } qw(real email uri);

    &add_stats($input{question}, $input{given}, $input{time}, $input{points}, $result );
    &add_entry(
	       time => $time,
	       points => $points,
	       user => $input{user},
	       date => scalar localtime(time),
	       %userdat
	      );
    $input{question}++; # use next one
  }
  return 1;
}





sub quest {
  if ($input{user} eq "") {
    &default;
    print "<p>Username required</p>\n";
    return;
  }

  # check given answer, if any, and print cookie+content-type
  return if(! &compute_question );

  if (exists $config{question}->{ $input{question} }) {
    my $T = new Template(-file => "$path_templates/question.tpl");
    my %question = %{ $config{question}->{ $input{question} } };

    if ($config{config}->{counter_time}) {
      $T->code(
	       timer => {
			 max      => $question{time},
			 question => $question{question},
			 size     => length($question{time})
			}
	      );
    }

    if ($config{config}->{counter_points}) {
      $T->code(
	       pointer => {
			   max      => $question{points},
			   interval => $question{points} / $question{time},
			   question => $question{question},
			   size     => length($question{points})
			  }
	      );
    }

    # prepare choices
    my @answers;
    if ($question{type} eq "R") {
      my %c = ( choice => "Answer:");
      push @answers, \%c;
    }
    else {
      foreach my $choice (sort {$a cmp $b } keys %{$question{choice}}) {
	my %c = (
		 choice => $choice,
		 answer => $question{choice}->{$choice}
		);
	push @answers, \%c;
      }
    }

    $T->code(
	     $choice{ $question{type} } => \@answers,
	     form => {
		      question => $input{question},
		      quest    => $question{question},
		      user     => $input{user},
		      cgi      => $cgi,
		      time     => time
		     }
	    );

    $T->code( data => { question => $input{question} } );

    print $T->process;
  }
  else {
    # no more questions, present the result+stats
    &stats_out(1);
}

sub stats_out {
  my($complete) = @_;
  if($input{override}) {
    $complete = 1;
  }
  my %e = &get_entry($input{user});
  my @stat = &get_stats;

  &dump(\@stat);
  &dump(\%e);

  # calculate percent and graphical bar
  my $maxpoints;
  foreach my $q (keys %{$config{question}}) {
    $max += $config{question}->{$q}->{points};
  }
  my $percent = int($e{points} / ($max / 100));
  my $bar = &get_bar($percent);

  $e{points} = "$e{points}/$max";

  # calculate correct answers
  my $numanswers = scalar keys %{$config{question}};
  my $correct = 0;
  foreach my $result (map { $_->{result} } @stat) {
    if ($result =~ /correct/) {
      $correct++;
    }
  }

  # cosmetics
  if ($e{uri}) {
    if ($e{uri} !~ /^http/) {
      $e{uri} = "http://$e{uri}";
    }
    $e{uri} = "<a href=\"$e{uri}\">$e{uri}</a>";
  }

  my %handover = (
	   data => { %e, percent => "${percent}% $bar", answers => "$correct/$numanswers" },
	   form => { cgi => $cgi }
		  );
  if ($complete) {
    $handover{shead} = {};
    $handover{stat}  = \@stat;
  }

  my $T = new Template(-file => "$path_templates/result.tpl");
  $T->code( %handover );
  print $T->process;
}
}

sub get_bar {
  my($percent) = @_;
  my $bar = "<img src=\"green.png\" height=\"10\" width=\"$percent\">"
    . "<img src=\"red.png\" height=\"10\" width=\"" . (100 - $percent)  . "\">";
  return $bar;
}





### helper subs ###
sub prepare {
  if (! -d "$path_data/$config{key}") {
    system "mkdir $path_data/$config{key}" and die "Could not prepare application!";
  }
  &lock;
  tie %db, "GDBM_File", "$path_data/$config{key}/results.db", &GDBM_WRCREAT, 0640
    or die "Could not open database $path_data/$config{key}/results.db: $!";
  tie %stats, "GDBM_File", "$path_data/$config{key}/stats.db", &GDBM_WRCREAT, 0640
    or die "Could not open database $path_data/$config{key}/stats.db: $!";
}

sub get_entry {
  my $user = shift;
  my $data = $db{$user};
  $data =~ s/%%/\//g; # demask delimiter
  my ($points, $time, $date, $real, $email, $uri) = split /\//, $data;
  my %entry = (
	       user   => $user,
	       date   => $date,
	       points => $points,
	       real   => $real,
	       email  => $email,
	       uri    => $uri,
	       time   => &int_to_time($time)
	      );
  return %entry;
}

sub add_entry {
  my %e = @_;
  foreach my $field (keys %e) {
    $e{$field} =~ s/\//%%/g; # mask delimiter
  }
  $db{ $e{user} } = "$e{points}/$e{time}/$e{date}/$e{real}/$e{email}/$e{uri}";
}

sub int_to_time {
  #
  # convert given seconds into
  # human readable hour value (YY:MM:SS)
  #
  my $seconds = shift;
  my $hours = int($seconds / 3600);
  my $min   = int(($seconds - ($hours * 3600)) / 60);
  my $secs  = $seconds - ($hours * 3600) - ($min * 60);
  return sprintf "%02d:%02d:%02d", $hours, $min, $secs;
}

sub get_md5 {
  my ($name, $src);
  if ($config{config}->{cookie} eq "user") {
    $src  = $input{user};
    $name = "$input{user}-quizzsession";
  }
  elsif ($config{config}->{cookie} eq "ip") {
    $src  = $ENV{REMOTE_ADDR};
    $name = "$ENV{REMOTE_ADDR}-quizzsession";
  }
  elsif ($config{config}->{cookie} eq "both") {
    $src  = $input{user} . $ENV{REMOTE_ADDR};
    $name = "$input{user}-$ENV{REMOTE_ADDR}-quizzsession";
  }
  my $md5 = md5_base64($src);
  $q = $config{quiz_name};
  $q =~ s/\s//g;
  $name .= "-$q";
  return ($name, $md5);
}

sub get_stats {
  # format:
  # 1:A,C:20:380/..
  # question given-answers needed-time points
  my @entries = split /\//, $stats{$input{user}};
  my (%stat, @stat);
  foreach my $entry (@entries) {
    my($question, $given, $time, $points, $result) = split /:/, $entry;
    $given =~ s/%%/:/g; # de-mask delimiter
    if (!$given) {
      $given = "-";
    }
    my %q = %{ $config{question}->{$question} };
    my ($choice, $color);
    my %correct = map { $_ => 1 } split /,/, $q{result};
    if ($q{type} ne "R") {
      my %g = map { $_ => 1 } split /,/, $given;
      $given = "";
      foreach my $answer (sort {$a cmp $b} keys %{$q{choice}}) {
	if (exists $correct{$answer}) {
	  $color = "green";
	}
	else {
	  $color = "red";
	}

	if (exists $g{$answer}) {
	  $given .= "<b>X</b><br>";
	}
	else {
	  $given .= "<b>-</b><br>";
	}

	$choice .= "<b>$answer</b> - <font color=$color>$q{choice}->{$answer}</font><br>";
      }
    }
    else {
      $choice = $q{result};
    }
    if ($result eq "correct") {
      $color = "green";
    }
    else {
      $color = "red";
    }
    $stat{$question} = {
			pos      => $question,
			question => $q{question},
			given    => $given,
			choice   => $choice,
			time     => $time,
			points   => "$points/$q{points}",
			result   => "<font color=$color>$result</font>"
			};
  }

  foreach my $question (sort {$a <=> $b} keys %stat) {
    push @stat, $stat{$question};
  }

  return @stat;
}

sub add_stats {
  # arguments:
  # ($question, $given, $time, $points, $result) = @_;
  # format:
  # 1:A,C:20:380/..
  # question given-answers needed-time points
  my $add  = join ":", @_;
  if (exists $stats{$input{user}}) {
    $add = "/" . $add;
  }
  $stats{$input{user}} .= $add;
}

sub dump {
  my $var = shift;
  print "<pre>\n" . Dumper($var) . "\n</pre>" if($config{debug});
}

sub lock {
  open LOCK, ">/tmp/$config{key}.lock" or die "Could not create the lock file: $!\n";
  flock(LOCK, LOCK_EX) or die;
}

sub unlock {
  close LOCK;
}

END {
  eval {
    &unlock;
  };
}
